#include "stdafx.h"
#include "nstest.h"
#include "ddeobj.h"

#include "nstesdoc.h"
#include "stdarg.h"

//	Serialization, runtime info, dynamic creation.
IMPLEMENT_SERIAL( CDDEConversation, CObject, 1 )

//	Static initialization.
CObList CDDEConversation::m_colConversations;

//	Construction
CDDEConversation::CDDEConversation()	{
	m_hConv = NULL;
	CommonConstruction();
}
CDDEConversation::CDDEConversation(HCONV hConv)	{
	m_hConv = hConv;
	CommonConstruction();
}

void CDDEConversation::CommonConstruction()	{
	//	Initialize everything but the conversation handle.
	m_rIndex = m_colConversations.AddTail(this);
	
	TRACE("Creating Covnersation %p\n", this);
}

//	Destruction.
CDDEConversation::~CDDEConversation()	{
	//	Remove ourselves from the list of running conversations.
	m_colConversations.RemoveAt(m_rIndex);
	
	TRACE("Destroying Conversation %p\n", this);
}

CNstestDoc *CDDEConversation::GetDocument()	{
	CDDEObject *pDDE = CDDEObject::ResolveConversation(this);
	if(pDDE == NULL)	{
		return(NULL);
	}
	return pDDE->GetDocument();
}

//	Proactive disconnection.
void CDDEConversation::DoDisconnect()	{
	//	Just disconnect from the current HCONV.
	DdeDisconnect(m_hConv);	
	GetDocument()->Display(ExplainError());
	CDDEObject::DDEDisconnect(this);
	delete this;
}

//	Informaitonal Lookup.
CDDEConversation *CDDEConversation::ResolveConversation(HCONV hConv)	{
	//	Loop through all conversations, and return the obect handling the
	//		conversation.
	POSITION rTraverse = m_colConversations.GetHeadPosition();
	CDDEConversation *pRetVal = NULL;
	while(rTraverse != NULL)	{
		pRetVal = (CDDEConversation *)m_colConversations.GetNext(rTraverse);
		if(pRetVal->m_hConv == hConv)	{
			break;
		}
		else	{
			pRetVal = NULL;
		}
	}
	
	return(pRetVal);
}

//	Utility
void CDDEConversation::ClientPassed(HSZ hszArgs, const char *pFormat, ...)	{
	//	Initialize variable number of argumetns.
	va_list VarList;
	va_start(VarList, pFormat);	
	
	int i_ArgNum = 0;
	char *pScan = (char *)pFormat;
	char *pExtract;
	
	//	Loop through the arguments we are going to parse.
	while(pScan && *pScan)	{
		//	What argument are we currently looking for?
		i_ArgNum++;
		pExtract = ExtractArgument(hszArgs, i_ArgNum);
	
		if(0 == strncmp(pScan, "DW", 2))	{
			//	DWORD.
			DWORD *pWord;
			pWord = va_arg(VarList, DWORD *);
			
			if(pExtract == NULL)	{
				*pWord = 0x0;
			}
			else	{
				sscanf(pExtract, "%lu", pWord);
			}
		}
		else if(0 == strncmp(pScan, "QCS", 3))	{
			//	A quoted CString
			CString *pCS = va_arg(VarList, CString *);
			
			if(pExtract == NULL)	{
				pCS->Empty();
			}
			else	{
				//	Extractions took off the leading and ending quotes.
				char *pCopy = pExtract;
				while(*pCopy)	{
					if(*pCopy == '\\' && *(pCopy + 1) == '\"')	{
						pCopy++;
					}
					
					*pCS += *pCopy;
					pCopy++;
				}
			}
		}
		else if(0 == strncmp(pScan, "CS", 2))	{
			//	A CString
			CString *pCS = va_arg(VarList, CString *);

			if(pExtract == NULL)	{
				pCS->Empty();
			}
			else	{
				*pCS = pExtract;
			}
		}
		else if(0 == strncmp(pScan, "BL", 2))	{
			//	A boolean
			TwoByteBool *pBool = va_arg(VarList, TwoByteBool *);
			
			if(pExtract == NULL)	{
				*pBool = FALSE;
			}
			else	{
				//	Compare for a TRUE or a FALSE
				if(0 == stricmp(pExtract, "TRUE"))	{
					*pBool = TRUE;
				}
				else	{
					*pBool = FALSE;
				}
			}
		}
		
		//	Go on to the next argument in our format string.
		pScan = SkipToNextArgument(pScan);
		
		//	Free the memory that was used during extraction.
		if(pExtract != NULL)	{
			delete pExtract;
		}
	}
	
	//	Done with variable number of arguments
	va_end(VarList);
}

HDDEDATA CDDEConversation::ServerReturns(const char *pFormat, ...)	{
	va_list VarList;
	va_start(VarList, pFormat);

	char *pTraverse = (char *)pFormat;
	char caNumpad[64];
	CString csBuffer;
	CString csRetval;
	
	while(*pTraverse)	{
		//	Erase temp data from our last pass.
		caNumpad[0] = '\0';
		csBuffer.Empty();
	
		//	Compare our current format to the known formats
		if(0 == strncmp(pTraverse, "DW", 2))	{
			//	A DWORD.
			DWORD *pWord = va_arg(VarList, DWORD *);
			
			if(pWord != NULL)	{
                //  See if we're to use hex or not.
                if(((CNstestApp *)AfxGetApp())->m_bHex)  {
                    sprintf(caNumpad, "%#lx", *pWord);
                }
                else    {
				    sprintf(caNumpad, "%lu", *pWord);
                }
				csRetval += caNumpad;
			}
		}
		else if(0 == strncmp(pTraverse, "CS", 2))	{
			//	A CString, not quoted
			CString *pCS = va_arg(VarList, CString *);
			
			if(pCS != NULL)	{
				csRetval += *pCS;
			}
		}
		else if(0 == strncmp(pTraverse, "QCS", 3))	{
			//	A quoted CString
			CString *pQCS = va_arg(VarList, CString *);
			
			if(pQCS != NULL)	{
				csRetval += '\"';
				
				//	Need to escape any '"' to '\"', literally.
				char *pConvert = (char *)(const char *)*pQCS;
				while(*pConvert != '\0')	{
					if(*pConvert == '\"')	{
						csRetval += '\\';
					}
					csRetval += *pConvert;
					pConvert++;
				}
				csRetval += '\"';
			}
		}
		else if(0 == strncmp(pTraverse, "BL", 2))	{
			//	A boolean
			TwoByteBool *pBool = va_arg(VarList, TwoByteBool *);
			
			if(pBool != NULL)	{
				if(*pBool != FALSE)	{
					csRetval += "TRUE";
				}
				else	{
					csRetval += "FALSE";
				}
			}
		}
		
		pTraverse = SkipToNextArgument(pTraverse);
		
		if(*pTraverse != '\0')	{
			csRetval += ',';
		}
	}

	//	Done with varargs.
	va_end(VarList);
	
	if(csRetval.IsEmpty())	{
		return(NULL);
	}

	HDDEDATA Final;
	Final = DdeCreateDataHandle(CDDEObject::m_dwidInst,
#ifdef _WIN32
        (unsigned char *)(const char *)csRetval,
#else
		(void *)(const char *)csRetval, 
#endif // _WIN32
 		csRetval.GetLength() + 1, 0,
 		CDDEObject::ResolveConversation(this)->GetServiceHSZ(),
		CF_TEXT, 0);
	GetDocument()->Display(ExplainError());

	//	THIS IS A HACK.
	if(strchr(pFormat, ',') == NULL)	{
		if(strcmp(pFormat, "BL") == 0)	{
			TwoByteBool bData = FALSE;
			if(csRetval == "TRUE")	{
				bData = TRUE;
			}
			DdeFreeDataHandle(Final);
			GetDocument()->Display(ExplainError());
			
			Final = DdeCreateDataHandle(CDDEObject::m_dwidInst,
				(unsigned char *)&bData, sizeof(TwoByteBool), 0,
				CDDEObject::ResolveConversation(this)->GetServiceHSZ(), CF_TEXT, 0);
			GetDocument()->Display(ExplainError());
		}
		else if(strcmp(pFormat, "DW") == 0)	{
			DWORD dwData;
			dwData = strtoul(csRetval, NULL, 0);
			DdeFreeDataHandle(Final);
			GetDocument()->Display(ExplainError());
			
			Final = DdeCreateDataHandle(CDDEObject::m_dwidInst,
				(unsigned char *)&dwData, sizeof(DWORD), 0,
				CDDEObject::ResolveConversation(this)->GetServiceHSZ(), CF_TEXT, 0);
			GetDocument()->Display(ExplainError());
		}
	}

	return(Final);
}

//	Handle DDE Pokes
HDDEDATA CDDEConversation::ServerPoke(HSZ hszTopic, HSZ hszItem, HDDEDATA hDataPoke)	{
	//	Switch on the topic we're handling.
	int iTopic = CDDEObject::EnumTopic(hszTopic);
	switch(iTopic)	{
	case CDDEObject::m_EndProgress:
		return(WWW_EndProgress(hszItem));
	case CDDEObject::m_SetProgressRange:
		return(WWW_SetProgressRange(hszItem));
	case CDDEObject::m_URLEcho:
		return(WWW_URLEcho(hszItem));
	case CDDEObject::m_ViewDocFile:
		return(WWW_ViewDocFile(hszItem));
	case CDDEObject::m_WindowChange:
		return(WWW_WindowChange(hszItem));
	}

	TRACE("Unknown poke received\n");
	return(DDE_FNOTPROCESSED);
}

HDDEDATA CDDEConversation::WWW_EndProgress(HSZ hszItem)	{
	DWORD dwTransactionID;
	
	ClientPassed(hszItem, "DW", &dwTransactionID);
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_EndProgress(%lu);", (const char *)ExplainPoke(DDEFACK),
		dwTransactionID);
	GetDocument()->Display(aBuf);

	return(DDEFACK);
}
HDDEDATA CDDEConversation::WWW_SetProgressRange(HSZ hszItem)	{
	DWORD dwTransactionID;
	DWORD dwMaximum;
	
	ClientPassed(hszItem, "DW,DW", &dwTransactionID, &dwMaximum);
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_SetProgressRange(%lu,%lu);", (const char *)ExplainPoke(DDEFACK),
		dwTransactionID, dwMaximum);
	GetDocument()->Display(aBuf);
	
	return(DDEFACK);
}
HDDEDATA CDDEConversation::WWW_URLEcho(HSZ hszItem)	{
	CString csURL;
	CString csMIMEType;
	DWORD dwWindowID;
	CString csReferrer;
	
	ClientPassed(hszItem, "QCS,QCS,DW,QCS", &csURL, &csMIMEType, &dwWindowID,
		&csReferrer);
		
	char aBuf[1024];
	sprintf(aBuf, "%s WWW_URLEcho(%s,%s,%lu,%s);", (const char *)ExplainPoke(DDEFACK),
		(const char *)csURL, (const char *)csMIMEType, dwWindowID, (const char *)csReferrer);
	GetDocument()->Display(aBuf);
		
	return(DDEFACK);
}
HDDEDATA CDDEConversation::WWW_ViewDocFile(HSZ hszItem)	{
	CString csFileName;
	CString csURL;
	CString csMIMEType;
	DWORD dwWindowID;
	
	ClientPassed(hszItem, "QCS,QCS,QCS,DW", &csFileName, &csURL, &csMIMEType, &dwWindowID);
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_ViewDocFile(%s,%s,%s,%lu);", (const char *)ExplainPoke(DDEFACK),
		(const char *)csFileName, (const char *)csURL, (const char *)csMIMEType,
		dwWindowID);
	GetDocument()->Display(aBuf);
	
	return(DDEFACK);
}
HDDEDATA CDDEConversation::WWW_WindowChange(HSZ hszItem)	{
	DWORD dwWindowID;
	DWORD dwWindowFlags;
	DWORD dwX;
	DWORD dwY;
	DWORD dwCX;
	DWORD dwCY;
	
	ClientPassed(hszItem, "DW,DW,DW,DW,DW,DW", &dwWindowID, &dwWindowFlags, &dwX,
		&dwY, &dwCX, &dwCY);
		
	char aBuf[512];
	sprintf(aBuf, "%s WWW_WindowChange(%lu,%lu,%lu,%lu,%lu,%lu);", (const char *)ExplainPoke(DDEFACK),
		dwWindowID, dwWindowFlags, dwX, dwY, dwCX, dwCY);
	GetDocument()->Display(aBuf);
		
	return(DDEFACK);
}

//	Handle DDE Requests
HDDEDATA CDDEConversation::ServerRequest(HSZ hszTopic, HSZ hszItem)	{
	//	Switch on the topic we're handling.
	int iTopic = CDDEObject::EnumTopic(hszTopic);
	switch(iTopic)	{
	case CDDEObject::m_Alert:
		return(WWW_Alert(hszItem));
	case CDDEObject::m_BeginProgress:
		return(WWW_BeginProgress(hszItem));
	case CDDEObject::m_MakingProgress:
		return(WWW_MakingProgress(hszItem));
	case CDDEObject::m_OpenURL:
		return(WWW_OpenURL(hszItem));
	case CDDEObject::m_QueryViewer:
		return(WWW_QueryViewer(hszItem));
	}

	TRACE("Unknown request received\n");
	return(NULL);
}

HDDEDATA CDDEConversation::WWW_Alert(HSZ hszItem)	{
	CString csMessage;
	DWORD dwType;
	DWORD dwButtons;
	DWORD dwAnswer = 0;
		
	ClientPassed(hszItem, "QCS,DW,DW", &csMessage, &dwType, &dwButtons);
	
	char aBuf[512];
	sprintf(aBuf, "%lu WWW_Alert(%s,%lu,%lu);", dwAnswer, (const char *)csMessage,
		dwType, dwButtons);
	GetDocument()->Display(aBuf);
	
	return(ServerReturns("DW", &dwAnswer));
}
HDDEDATA CDDEConversation::WWW_BeginProgress(HSZ hszItem)	{
	DWORD dwWindowID;
	CString csInitialMessage;
	DWORD dwTransactionID = (DWORD)time(NULL);	
	
	ClientPassed(hszItem, "DW,QCS", &dwWindowID, &csInitialMessage);
	
	char aBuf[512];
	sprintf(aBuf, "%lu WWW_BeginProgress(%lu,%s);", dwTransactionID, dwWindowID,
		(const char *)csInitialMessage);
	GetDocument()->Display(aBuf);
	
	return(ServerReturns("DW", &dwTransactionID));
}
HDDEDATA CDDEConversation::WWW_MakingProgress(HSZ hszItem)	{
	DWORD dwTransactionID;
	CString csMessage;
	DWORD dwProgress;
	TwoByteBool bStop = FALSE;
	
	ClientPassed(hszItem, "DW,QCS,DW", &dwTransactionID, &csMessage, &dwProgress);
	
	char aBuf[512];
	sprintf(aBuf, "%s WWW_MakingProgress(%lu,%s,%lu);", (const char *)ExplainBool(bStop),
		dwTransactionID, (const char *)csMessage, dwProgress);
	GetDocument()->Display(aBuf);
	
	return(ServerReturns("BL", &bStop));
}
HDDEDATA CDDEConversation::WWW_OpenURL(HSZ hszItem)	{
	CString csURL;
	CString csSaveAs;
	DWORD dwWindowID;
	DWORD dwFlags;
	CString csPostFormData;
	CString csPostMIMEType;
	CString csProgressServer;
	DWORD dwServicingID = (DWORD)time(NULL);
	
	ClientPassed(hszItem, "QCS,QCS,DW,DW,QCS,QCS,CS", &csURL, &csSaveAs,
		&dwWindowID, &dwFlags, &csPostFormData, &csPostMIMEType, &csProgressServer);
		
	char aBuf[2048];
	sprintf(aBuf, "%lu WWW_OpenURL(%s,%s,%lu,%lu,%s,%s,%s);", dwServicingID,
		(const char *)csURL, (const char *)csSaveAs, dwWindowID, dwFlags,
		(const char *)csPostFormData, (const char *)csPostMIMEType,
		(const char *)csProgressServer);
	GetDocument()->Display(aBuf);
		
	return(ServerReturns("DW", &dwServicingID));
}
HDDEDATA CDDEConversation::WWW_QueryViewer(HSZ hszItem)	{
	CString csURL;
	CString csMIMEType;
	char *pTemp = _tempnam("c:\\temp", "nstest");
	CString csSaveIn = (pTemp == NULL) ? "\\delme.nst" : pTemp;
	
	ClientPassed(hszItem, "QCS,QCS", &csURL, &csMIMEType);
	
	char aBuf[2048];
	sprintf(aBuf, "%s WWW_QueryViewer(%s,%s);", (const char *)csSaveIn,
		(const char *)csURL, (const char *)csMIMEType);
	GetDocument()->Display(aBuf);
	
	return(ServerReturns("QCS", &csSaveIn));
}


//	Client Requests and pokes.
HDDEDATA CDDEConversation::ClientRequest(HSZ hszItem)	{
	//	Send along the item to the server, as an XTYP_REQUEST.
	//	Return the data received from the server.
	//	Proactively delete the hszItem string.
	HDDEDATA hRetVal = DdeClientTransaction(NULL, 0ul, m_hConv, hszItem, CF_TEXT, XTYP_REQUEST, DDETIMEOUT, NULL);
	GetDocument()->Display(ExplainError());	
	if(hszItem != NULL)	{
		DdeFreeStringHandle(CDDEObject::m_dwidInst, hszItem);
		GetDocument()->Display(ExplainError());
	}
	return(hRetVal);
}
HDDEDATA CDDEConversation::ClientPoke(HSZ hszItem)	{
	//	Send along the item to the server, as an XTYP_POKE.
	//	Return the data received from the server.
	//	Proactively delete the hszItem string.
	HDDEDATA hRetVal = DdeClientTransaction(NULL, 0ul, m_hConv, hszItem, CF_TEXT, XTYP_POKE, DDETIMEOUT, NULL);
	GetDocument()->Display(ExplainError());
	if(hszItem != NULL)	{
		DdeFreeStringHandle(CDDEObject::m_dwidInst, hszItem);
		GetDocument()->Display(ExplainError());
	}
	return(hRetVal);
}

//	Persistance
void CDDEConversation::Serialize(CArchive& ar)	{
	if(ar.IsStoring())	{
		//	Add storing code
	}
	else	{
		//	Add loading code
	}
	
	//	Serialize the base.
	CObject::Serialize(ar);
}

#ifdef _DEBUG

void CDDEConversation::AssertValid() const	{
	//	Call base to assert valid.
	CObject::AssertValid();
	
	//	Check our instance specific members.
	ASSERT(m_hConv);	//	Better have a conversation!
	ASSERT(m_rIndex);	//	Position in list.
}

void CDDEConversation::Dump(CDumpContext& dc) const	{
	//	Have the base dump
	CObject::Dump(dc);
	
	//	Now we dump any relevant information....
}

#endif // _DEBUG
